/***********************************************************************
*
* sim990.c - Simulator for the TI 990 computer.
*
* Changes:
*   05/29/03   DGP   Original.
*   06/20/03   DGP   Added interrupt support.
*
***********************************************************************/

#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include <signal.h>
#include <errno.h>

#include "simdef.h"
#include "screen.h"

uint16 pcreg;		/* The program PC */
uint16 statreg;		/* The program status register */
uint16 wpreg;		/* The program Workspace Pointer */
uint16 lights;		/* The panel lights */
uint16 breakaddr;	/* Break point address */
long instcount;		/* Instructions executed */
long pancount;		/* Instructions executed for panel */

int run = FALSE;	/* CPU running */
int idle = FALSE;	/* CPU running IDLE or JXX $ */
int fault = FALSE;	/* CPU fault */
int enablepanel = FALSE; /* Enable panel */
int devcnt = 0;		/* Device attached count */
int model = 4;		/* Which model */
int breakflag = FALSE;	/* No breakpoint enabled */
int traceflag = FALSE;	/* No trace enabled */
int deferint = 0;	/* Defer interrupts inst count */
char romfile[256];	/* ROM file name */
char view[MAXVIEW][81];	/* Messages panel */
uint8 memory[MEMSIZE];	/* The CPU memory */
Device devices[MAXDEVICES]; /* The attached devices */
FILE *tracefd = NULL;	/* Trace file fd */

extern int errno;

static uint16 disp = 0;	/* Display area address */

/***********************************************************************
* sigintr - Process interrupt signal.
***********************************************************************/

void
sigintr (int sig)
{
   run = FALSE;
   signal (SIGINT, sigintr);
}

/***********************************************************************
* puthelp - Put help on screen.
***********************************************************************/

static void
puthelp(void)
{
   screenposition (14,1);
   clearline();
   printf ("  a - attach  b - boot    c - clrBrk  d - display g - go");
   screenposition (15,1);
   clearline();
   printf ("  h - help    k - break   l - load    m - modify  n - next");
   screenposition (16,1);
   clearline();
   printf ("  p - set pc  q - quit    r - reset   s - step    w - set wp");
   screenposition (17,1);
   clearline();
   printf ("  z - panel");
   screenposition (18,1);
   clearline();
   printf ("  ");
}

/***********************************************************************
* putview - Puts messages on screen.
***********************************************************************/

static void
putview (void)
{
   int i;

   for (i = 0; i < MAXVIEW; i++)
   {
      if (!enablepanel && view[i][0] == '\0') return;
      screenposition (14+i,1);
      clearline();
      if (view[i][0]) printf ("%s", view[i]);
      view[i][0] = '\0';
   }
}

/***********************************************************************
* getnum - Get a number from the command.
***********************************************************************/

char *
getnum (char *bp, int *val)
{
   char *cp;
   int i = 0;

   while (isspace(*bp)) bp++;
   cp = bp;
   if (*bp == HEXSYM)
   {
      i = strtol (++bp, &cp, 16);
   }
   else if (isdigit(*bp) || *bp == '-' || *bp == '+')
   {
      i = strtol (bp, &cp, 10);
   }
   *val = i;
   return (cp);
}

/***********************************************************************
* panel - Put up the panel.
***********************************************************************/

void
panel (void)
{
   char c1, c2;

   c1 = DISPMEM >> 8;
   c2 = DISPMEM & 0xFF;
   if (!isprint(c1)) c1 = '.';
   if (!isprint(c2)) c2 = '.';

   screenposition (1,1);
   clearline();
   printf ("%s %s      Model 990/%d ", INTRO, VERSION, model);

   screenposition (3,1);
   printf (" PC  %04X  WP  %04X  ST  %04X  FP  %04X",
	    pcreg, wpreg, statreg, lights);

   screenposition (5,1);
   printf (" DISP  %04X %04X %c%c  %-5.5s  %-5.5s  %-5.5s",
	    disp & 0xFFFF, DISPMEM, c1,c2,
	    run ? "RUN" : " ",
	    idle ? "IDLE" : " ",
	    fault ? "FAULT" : " ");

   screenposition (6,1);
   printf (" BREAK %04X        COUNT %ld",
	    breakaddr,
	    instcount);

   screenposition (8,1);
   printf ("Workspace:");
   screenposition (9,1);
   printf (" R00 %04X  R01 %04X  R02 %04X  R03 %04X", R0, R1, R2, R3);
   screenposition (10,1);
   printf (" R04 %04X  R05 %04X  R06 %04X  R07 %04X", R4, R5, R6, R7);
   screenposition (11,1);
   printf (" R08 %04X  R09 %04X  R10 %04X  R11 %04X", R8, R9, R10, R11);
   screenposition (12,1);
   printf (" R12 %04X  R13 %04X  R14 %04X  R15 %04X", R12, R13, R14, R15);
   screenposition (14,1);
   fflush(stdout);
   pancount = FALSE;

}

/***********************************************************************
* clearall - Clear out system.
***********************************************************************/

static int 
clearall()
{
   FILE *rom;
   int i;
   uint16 lp;
   uint16 val;
   char filename[256];

   /*
   ** Clear registers
   */

   pcreg = 0;
   statreg = 0;
   wpreg = 0;
   lights = 0;

   /*
   ** Clear memory
   */

   memset (&memory, '\0', sizeof(memory));
   
   /*
   ** Load ROM
   */

   if (romfile[0])
   {
      strcpy (filename, romfile);
   }
   else
   {
      sprintf (filename, "ti990%d.rom", model);
   }

   if ((rom = fopen (filename, "r")) == NULL)
   {
      sprintf (view[0], "clearall: ROM load failed: %s",
	    strerror(errno));
      sprintf (view[1], "filename: %s", filename);
      return (-1);;
   }

   lp = 0xF000;

   for (i = 0; i < 4096; i+=2)
   {
      int c;

      if ((c = fgetc (rom)) == EOF)
      {
	 sprintf (view[0], "clearall: ROM read failed: %s",
	       strerror(errno));
	 sprintf (view[1], "filename: %s", filename);
	 return (-1);;
      }
      val = (c & 0xFF) << 8;
      if ((c = fgetc (rom)) == EOF)
      {
	 sprintf (view[0], "clearall: ROM read failed: %s",
	       strerror(errno));
	 sprintf (view[1], "filename: %s", filename);
	 return (-1);;
      }
      val = val | (c & 0xFF);
      memory[lp+MSB] = (val) >> 8 & 0xFF;
      memory[lp+LSB] = (val) & 0xFF;
      lp += 2;
   }
   fclose (rom);

   instcount = 0;
   clearscreen();

   return (0);
}

/***********************************************************************
* main - Main Procedure.
***********************************************************************/

main (int argc, char **argv)
{
   char *bp;
   int action;
   int i;
   int norun;
   char input[80];
   char oinput[80];

   /*
   ** Process args
   */

   for (i = 1; i < argc; i++)
   {
      bp = argv[i];

      if (*bp == '-')
      {
         bp++;
         switch (*bp)
         {
	 case 'm':
	    i++;
	    model = atoi (argv[i]);
	    break;

         case 'p':
            enablepanel = TRUE;
            break;

	 case 'r':
	    i++;
	    strcpy (romfile, argv[i]);
	    break;

         default:
      USAGE:
	    printf ("usage:  sim990 [-m model] [-p] [-r romfile]\n");
	    return (ABORT);
         }
      }
   }

   /*
   ** Clear system
   */

   clearall();

   /*
   ** Set up initial devices
   */

   for (i = 0; i < MAXDEVICES; i++)
   {
      memset (&devices[i], '\0', sizeof (Device));
   }
   devices[0].type = CONIN;
   devices[0].fd = stdin;
   devices[0].devaddr = TERMINAL;
   devices[0].cdevaddr = 0;
   devices[0].intlvl = TERMINT;
   strcpy (devices[0].dev, "CI");
   strcpy (devices[0].file, "stdin");

   devices[1].type = CONOUT;
   devices[1].fd = stdout;
   devices[1].devaddr = TERMINAL;
   devices[1].cdevaddr = 0;
   devices[1].intlvl = TERMINT;
   strcpy (devices[1].dev, "CO");
   strcpy (devices[1].file, "stdout");
   devcnt = 2;

   ttinit ();

   /*
   ** Fire up clock thread
   */
   startclk();

   /*
   ** stop the interrupts
   */

   stopclk();


   if (!enablepanel)
      printf ("%s %s: Model 990/%d\n", INTRO, VERSION, model);

   input[0] = oinput[0] = '\0';
   action = NULL_ACTION;

   while (action != QUIT_ACTION)
   {
      if (enablepanel) panel();

      if (view[0][0] != '\0') putview ();

      screenposition (23,1);
      clearline();
      if (enablepanel)
      {
	 printf (":%s", oinput);
	 screenposition (23,2);
      }
      else
         printf (":");
      fgets (input, sizeof(input), stdin);
      *(strchr(input,'\n')) = '\0';
      if (input[0] == '\0')
         strcpy (input, oinput);
      else
	 strcpy (oinput, input);
      putview ();

      bp = input;
      switch (*bp++)
      {
      case 'a': /* Attach device */
	 if (*bp)
	    attachdev (bp);
	 if (view[0][0] == '\0')
	 {
	    sprintf (view[0], "DEV  ADDR INT  FILE");
	    for (i = 0; i < devcnt; i++)
	    {
	       if (devices[i].fd)
		  sprintf (view[i+1], "%s   %04X %2d  %s",
		     devices[i].dev, devices[i].devaddr,
		     devices[i].intlvl, devices[i].file);
	    }
	 }
	 action = NULL_ACTION;
	 oinput[0] = '\0';
	 break;

      case 'b': /* Boot */
	 if (*bp == 'n') /* No start option */
	 {
	    bp++;
	    norun = TRUE;
	 }
	 else norun = FALSE;
	 action = NULL_ACTION;
	 if (!*bp)
	    strcpy (bp, "f0");

	 if (boot (bp) == 0)
	 {
	    if (!norun && run)
	    {
	       action = RUN_ACTION;
	       signal (SIGINT, sigintr);
	       runprog();
	       signal (SIGINT, SIG_DFL);
	    }
	 }
	 oinput[0] = '\0';
         break;

      case 'c': /* Clear breakpoint */
	 breakflag = FALSE;
	 action = NULL_ACTION;
	 break;

      case 'd': /* Display memory */
	 bp = getnum (bp, &action);
	 disp = action;
	 action = DISP_ACTION;
	 if (!enablepanel)
	    printf (" (%04X) = %04X\n", disp & 0xFFFF, DISPMEM);
	 oinput[0] = '\0';
	 break;

      case 'g': /* Go - run program */
	 action = RUN_ACTION;
	 run = TRUE;
	 signal (SIGINT, sigintr);
	 runprog();
	 signal (SIGINT, SIG_DFL);
	 oinput[0] = '\0';
	 break;

      case 'h': /* Help */
	 puthelp();
	 action = HELP_ACTION;
	 oinput[0] = '\0';
	 break;

      case 'k': /* set breaKpoint */
         bp = getnum (bp, &action);
	 breakaddr = action;
	 breakflag = TRUE;
	 action = NULL_ACTION;
	 oinput[0] = '\0';
	 break;

      case 'l': /* Load - use binloader */
	 if (*bp == 'n') /* No start option */
	 {
	    bp++;
	    norun = TRUE;
	 }
	 else norun = FALSE;
	 bp = getnum (bp, &action);
	 if (binloader (bp, action) == 0)
	 {
	    if (!norun && run)
	    {
	       action = RUN_ACTION;
	       signal (SIGINT, sigintr);
	       runprog();
	       signal (SIGINT, SIG_DFL);
	    }
	    else action = LOAD_ACTION;
	 }
	 run = FALSE;
	 oinput[0] = '\0';
	 break;

      case 'm': /* Modify memory */
	 bp = getnum (bp, &action);
	 PUTMEM (disp, action);
	 action = DISP_ACTION;
	 if (!enablepanel)
	    printf (" (%04X) = %04X\n", disp & 0xFFFF, DISPMEM);
	 oinput[0] = '\0';
	 break;

      case 'n': /* display Next memory location */
	 disp += 2;
	 action = DISP_ACTION;
	 if (!enablepanel)
	    printf (" (%04X) = %04X\n", disp & 0xFFFF, DISPMEM);
	 break;

      case 'p': /* set PC */
	 bp = getnum (bp, &action);
	 pcreg = action;
	 action = PC_ACTION;
	 oinput[0] = '\0';
	 break;

      case 'q': /* Quit */
	 action = QUIT_ACTION;
	 break;

      case 'r': /* Reset system */
         clearall();
	 action = CLEAR_ACTION;
	 oinput[0] = '\0';
	 break;

      case 's': /* Step program */
	 action = STEP_ACTION;
	 run = FALSE;
	 stepprog();
	 break;

      case 't': /* Trace program */
	 traceflag = traceflag ? FALSE : TRUE;
	 if (traceflag)
	 {
	    while (isspace(*bp)) bp++;
	    if ((tracefd = fopen (bp, "w")) == NULL)
	    {
	       traceflag = FALSE;
	       perror ("Can't open trace file");
	    }
	 }
	 else
	 {
	    fclose (tracefd);
	    tracefd = NULL;
	 }
	 break;

      case 'w': /* set WP */
	 bp = getnum (bp, &action);
	 wpreg = action;
	 action = WP_ACTION;
	 oinput[0] = '\0';
	 break;

      case 'z': /* it's Zee panel */
	 action = NULL_ACTION;
	 enablepanel = enablepanel ? FALSE : TRUE;
	 if (enablepanel)
	 {
	    clearscreen();
	    panel();
	 }
	 break;

      default: puthelp();
      }

   }

   /*
   ** Close open devices
   */

   for (i = 0; i < devcnt; i++)
   {
      if (devices[i].fd != NULL)
      {
         if (!(devices[i].fd == stdin || devices[i].fd == stdout))
	    fclose (devices[i].fd);
      }
   }

   ttclose ();
   screenposition (25,1);
}
