/***********************************************************************
*
* simio.c - IO processing for the TI 990 Simulator.
*
* Changes:
*   05/29/03   DGP   Original.
*   06/20/03   DGP   Added interrupt support.
*   07/01/03   DGP   Fixed floppy unit on STCR of status.
*   07/08/03   DGP   Changed FLOPPY to return status if unit not present.
*   07/09/03   DGP   Forced PC and WP to be on even boundry for interrupts.
*   07/15/03   DGP   Added Card Reader support.
*
***********************************************************************/

#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include <errno.h>
#if defined (_WIN32)
#define __TTYROUTINES 0
#include <conio.h>
#include <windows.h>
#include <signal.h>
#endif
#if defined(UNIX)
#include <sys/time.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <pthread.h>
#endif

#include "simdef.h"
#include "simcra.h"

extern uint16 pcreg;	/* The program PC */
extern uint16 statreg;	/* The program status register */
extern uint16 wpreg;	/* The program Workspace Pointer */
extern uint16 lights;	/* The panel lights */

extern int run;
extern int idle;
extern int fault;
extern int devcnt;
extern int model;
extern int pancount;
extern char view[MAXVIEW][81];

extern uint8 memory[MEMSIZE];
extern Device devices[MAXDEVICES];

extern int errno;

static int clockactive = FALSE;
static int clockthreadactive = FALSE;
static uint16 intrequest = 0;
static int sim_int_char = 005;

/*
** Terminal routines borrowed from Bub Supnik's simh
*/

#if defined (_WIN32)
static volatile int sim_win_ctlc = 0;

void win_handler (int sig)
{
sim_win_ctlc = 1;
return;
}

int ttinit (void)
{
return 0;
}

int ttrunstate (void)
{
if (signal (SIGINT, win_handler) == SIG_ERR) return -1;
SetThreadPriority (GetCurrentThread(), THREAD_PRIORITY_BELOW_NORMAL);
return 0;
}

int ttcmdstate (void)
{
SetThreadPriority (GetCurrentThread(), THREAD_PRIORITY_NORMAL);
return 0;
}

int ttclose (void)
{
return 0;
}

int sim_os_poll_kbd (void)
{
   int c;

   if (sim_win_ctlc)
   {
	sim_win_ctlc = 0;
	signal (SIGINT, win_handler);
	return 003;
   }
   if (!kbhit ()) return EOF;
   c = _getch ();
   if ((c & 0177) == '\b') c = 0177;
   if ((c & 0177) == sim_int_char)
   {
      run = FALSE;
      return EOF;
   }
   if (islower(c)) c = toupper(c);
   return c;
}

int sim_os_putchar (int32 c)
{
if (c != 0177)
	_putch (c & 0x7F);
return 0;
}
#endif

#if defined(UNIX)
struct termios cmdtty, runtty;

int ttinit (void)
{
if (!isatty (fileno (stdin))) return 0;			/* skip if !tty */
if (tcgetattr (0, &cmdtty) < 0) return -1;		/* get old flags */
runtty = cmdtty;
runtty.c_lflag = runtty.c_lflag & ~(ECHO | ICANON);	/* no echo or edit */
runtty.c_oflag = runtty.c_oflag & ~OPOST;		/* no output edit */
runtty.c_iflag = runtty.c_iflag & ~ICRNL;		/* no cr conversion */
runtty.c_cc[VINTR] = sim_int_char;			/* interrupt */
runtty.c_cc[VQUIT] = 0;					/* no quit */
runtty.c_cc[VERASE] = 0;
runtty.c_cc[VKILL] = 0;
runtty.c_cc[VEOF] = 0;
runtty.c_cc[VEOL] = 0;
runtty.c_cc[VSTART] = 0;				/* no host sync */
runtty.c_cc[VSUSP] = 0;
runtty.c_cc[VSTOP] = 0;
#if defined (VREPRINT)
runtty.c_cc[VREPRINT] = 0;				/* no specials */
#endif
#if defined (VDISCARD)
runtty.c_cc[VDISCARD] = 0;
#endif
#if defined (VWERASE)
runtty.c_cc[VWERASE] = 0;
#endif
#if defined (VLNEXT)
runtty.c_cc[VLNEXT] = 0;
#endif
runtty.c_cc[VMIN] = 0;					/* no waiting */
runtty.c_cc[VTIME] = 0;
#if defined (VDSUSP)
runtty.c_cc[VDSUSP] = 0;
#endif
#if defined (VSTATUS)
runtty.c_cc[VSTATUS] = 0;
#endif
return 0;
}

int ttrunstate (void)
{
if (!isatty (fileno (stdin))) return 0;			/* skip if !tty */
runtty.c_cc[VINTR] = sim_int_char;			/* in case changed */
if (tcsetattr (0, TCSAFLUSH, &runtty) < 0) return -1;
return 0;
}

int ttcmdstate (void)
{
if (!isatty (fileno (stdin))) return 0;			/* skip if !tty */
if (tcsetattr (0, TCSAFLUSH, &cmdtty) < 0) return -1;
return 0;
}

int ttclose (void)
{
return ttcmdstate ();
}

int sim_os_poll_kbd (void)
{
int status;
unsigned char buf[1];

status = read (0, buf, 1);
if (status != 1) return EOF;
if (islower (buf[0])) buf[0] = toupper(buf[0]);
return (buf[0]);
}

int sim_os_putchar (uint8 out)
{
char c;

c = out & 0x7F;
write (1, &c, 1);
return 0;
}
#endif

/***************************************************************************
* smallsleep - sleep in U seconds
***************************************************************************/

static void
smallsleep (long Usec)
{
#if defined(UNIX)
   struct timeval tv;

   tv.tv_sec = 0;
   tv.tv_usec = Usec;
   if (select (1,
	       NULL,
	       NULL,
	       NULL,
	       &tv) < 0)
   {
      if (errno != EINTR && errno != EAGAIN)
	 perror ("smallsleep: select failed");
   }
#endif /* UNIX */

#if defined(_WIN32)
   Sleep ((unsigned long) (Usec) / 100L);
#endif /* _WIN32 */

#if defined(OS2)
   DosSleep ((unsigned long) (Usec) / 100L);
#endif /* OS2 */

}

/***********************************************************************
* clock_thread - run the system clock
***********************************************************************/

static void
clock_thread (void)
{
   while (TRUE)
   {
      smallsleep (10000);
      if (clockactive)
	 intrequest = intrequest | (1 << CLOCKINT);
      pancount = TRUE;
   }
}

/***********************************************************************
* finddev - Find attached device.
***********************************************************************/

static int
finddev (uint16 devaddr)
{
   int i;

   for (i = 0; i < devcnt; i++)
   {
      if (devices[i].devaddr == devaddr && devices[i].fd != NULL) break;
      if (devices[i].cdevaddr == devaddr && devices[i].fd != NULL) break;
   }

   if (i == devcnt)
   {
#if 0
      run = FALSE;
      sprintf (view[0], "Device for devaddr %04X is not attached", devaddr);
#endif
      return (-1);
   }
   return (i);
}

/***********************************************************************
* findtype - Find attached device by type.
***********************************************************************/

static int
findtype (int type, char *dt)
{
   int i;

   for (i = 0; i < devcnt; i++)
   {
      if (devices[i].type == type && devices[i].fd != NULL &&
	  !strcmp (devices[i].dev, dt) ) return (i);
   }

   return (-1);
}

/***********************************************************************
* attachit - Attach a file to a device (do the dirty work).
***********************************************************************/

static void
attachit (char *bp, int type, char *mode, char *dev)
{
   FILE *nfd;
   int new;
   int i;
   int devaddr;
   int interrupt;

   bp = getnum (bp, &devaddr);
   bp = getnum (bp, &interrupt);
   if (interrupt > 0)
   {
      if (model == 4 && (interrupt < 2 || interrupt > 7))
      {
	 sprintf (view[0], "attach: illegal interrupt %d for 990/%d",
		  interrupt, model);
	 return;
      }
   }

   while (isspace(*bp)) bp++;
   if ((nfd = fopen (bp, mode)) == NULL)
   {
      sprintf (view[0], "attach: open failed: %s",
	       strerror(errno));
      sprintf (view[1], "filename: %s", bp);
      return;
   }

   if ((i = findtype (type, dev)) >= 0)
   {
      if (!(devices[i].fd == stdin || devices[i].fd == stdout))
	 fclose (devices[i].fd);
   }
   else
   {
      i = devcnt;
      devcnt++;
   }
   devices[i].fd = nfd;
   devices[i].devaddr = devaddr;
   devices[i].type = type;
   devices[i].select = 0;
   devices[i].laststat = 0;
   devices[i].count = 0;
   devices[i].unit = 0;
   devices[i].intlvl = interrupt;
   if (type == FLOPPY) 
   {
      devices[i].cdevaddr = devaddr + 0x20;
      devices[i].unit = dev[1] - '0';
#ifdef DEBUGFLOPPY
      fprintf (stderr, "Attach floppy, unit = %d, cru = %04x to %s\n",
	       devices[i].unit, devices[i].devaddr, bp);
#endif
   }
   strcpy (devices[i].file, bp);
   strcpy (devices[i].dev, dev);
}

/***********************************************************************
* attachdev - Attach a file to a device.
***********************************************************************/

void
attachdev (char *bp)
{
   char fdev[10];

   while (isspace (*bp)) bp++;
   switch (*bp++)
   {
      case 'c':
	 switch (*bp++)
	 {
	 case 'i': /* Console input */
	    attachit (bp, CONIN, "r", "CI");
	    break;

	 case 'o': /* Console output */
	    attachit (bp, CONOUT, "w", "CO");
	    break;

	 case 'r': /* Card reader */
	    attachit (bp, CARDREADER, "r", "CR");
	    break;

	 default: 
	    goto BADDEV;
	 }
         break;

      case 'f':
	 if (!isdigit (*bp)) goto BADDEV;
	 sprintf (fdev, "F%c", *bp++);
	 attachit (bp, FLOPPY, "r+", fdev);
         break;

      case 'p':
	 attachit (bp, PRINTER, "w", "PR");
         break;

      case 's':
	 attachit (bp, SERIAL, "r+", "SE");
         break;

      default:
BADDEV:
         sprintf (view[0], "Invalid device %s, devices are:", bp);
	 sprintf (view[1], " ci = console input  co = console output");
	 sprintf (view[2], " p  = printer        fn = floppy n=0-3");
	 sprintf (view[3], " cr = card reader");
	 sprintf (view[4], "usage: a device devaddr intlvl file");
	 break;
   }
}

/***********************************************************************
* setbitone - Set CRU bit to one.
***********************************************************************/

void
setbitone (uint16 inst)
{
   Device *dev;
   int i;
   uint16 devaddr;
   uint8 disp;

   disp = inst & 0x00FF;
   devaddr = R12;

#ifdef DEBUGCRU
   fprintf (stderr, "CRU %04X SBO %d\n", devaddr, disp);
#endif

   /*
   ** Front panel is always attached
   */

   if (devaddr == PANEL)
   {
      if (disp == 11) fault = TRUE;
      return;
   }
   
   if ((devaddr >> 8) == 0x17) /* /5 onboard ports */
   {
      return;
   }

   if ((i = finddev (devaddr)) < 0) return;
   dev = &devices[i];
   switch (dev->type)
   {
   case CONIN:
   case CONOUT:
      if (disp == 14)
      {
         i = findtype (CONIN, "CI");
	 dev = &devices[i];
	 dev->intenabled = TRUE;
	 intrequest = intrequest | (1 << dev->intlvl);
      }
      break;
   case SERIAL:
      if (disp == 14)
      {
	 if (dev->intlvl > 0)
	 {
	    dev->intenabled = TRUE;
	    intrequest = intrequest | (1 << dev->intlvl);
	 }
      }
      break;
#ifdef DEBUGFLOPPY
   case FLOPPY:
      fprintf (stderr, "CRU %04X SBO %d\n", devaddr, disp);
      break;
#endif
#ifdef DEBUGPRINTER
   case PRINTER:
      fprintf (stderr, "CRU %04X SBO %d\n", devaddr, disp);
      break;
#endif
   case CARDREADER:
#ifdef DEBUGCARD
      fprintf (stderr, "CRU %04X SBO %d\n", devaddr, disp);
#endif
      if (disp == 13)
      {
	 dev->intenabled = TRUE;
	 dev->select = TRUE;
	 intrequest = intrequest | (1 << dev->intlvl);
      }
      break;
   default: ;
   }
   return;
}

/***********************************************************************
* setbitzero - Set CRU bit to zero.
***********************************************************************/

void
setbitzero (uint16 inst)
{
   Device *dev;
   int i;
   uint16 devaddr;
   uint8 disp;

   disp = inst & 0x00FF;
   devaddr = R12;

#ifdef DEBUGCRU
   fprintf (stderr, "CRU %04X SBZ %d\n", devaddr, disp);
#endif

   /*
   ** Front panel is always attached
   */

   if (devaddr == PANEL)
   {
      if (disp == 11) fault = FALSE;
      return;
   }
   
   if ((devaddr >> 8) == 0x17) /* /5 onboard ports */
   {
      return;
   }

   if ((i = finddev (devaddr)) < 0) return;
   dev = &devices[i];
#ifdef DEBUGFLOPPY
   if (dev->type == FLOPPY)
      fprintf (stderr, "CRU %04X SBZ %d\n", devaddr, disp);
#endif
#ifdef DEBUGPRINTER
   if (dev->type == PRINTER)
      fprintf (stderr, "CRU %04X SBZ %d\n", devaddr, disp);
#endif
   if (dev->type == CARDREADER)
   {
#ifdef DEBUGCARD
      fprintf (stderr, "CRU %04X SBZ %d\n", devaddr, disp);
#endif
      switch (disp)
      {
      case 9:
	 if (dev->select)
	 {
	    dev->intenabled = TRUE;
	    intrequest = intrequest | (1 << dev->intlvl);
	 }
	 break;
      case 13:
	 dev->intenabled = FALSE;
	 dev->select = FALSE;
	 intrequest = intrequest | (1 << dev->intlvl);
	 break;
      case 14:
	 if (dev->select)
	 {
	    dev->intenabled = TRUE;
	    intrequest = intrequest | (1 << dev->intlvl);
	 }
	 break;
      default: ;
      }
   }

   return;
}

/***********************************************************************
* testbit - Test CRU bit.
***********************************************************************/

void
testbit (uint16 inst)
{
   Device *dev;
   int i;
   uint16 devaddr;
   uint8 disp;

   disp = inst & 0x00FF;
   devaddr = R12;

#ifdef DEBUGCRU
   fprintf (stderr, "CRU %04X TB %d\n", devaddr, disp);
#endif

   /*
   ** Front panel is always attached
   */

   if (devaddr == PANEL)
   {
      CLR_EQ;
      return;
   }
   
   if ((devaddr >> 8) == 0x17) /* /5 onboard ports */
   {
      return;
   }

   if ((i = finddev (devaddr)) < 0) return;
   dev = &devices[i];

   if (dev->type == CONIN || dev->type == CONOUT)
   {
      switch (disp)
      {
      case 0:  /* TTY */
      case 12: /* RRQ */
         i = findtype (CONIN, "CI");
	 break;
      case 8: /* XMIT */
      case 11: /* WRQ */
      case 13: /* DCD */
      case 14: /* DSR */
      default:
         i = findtype (CONOUT, "CO");
      }
      dev = &devices[i];
   }

   switch (dev->type)
   {
   case CONIN:
      if (disp == 0) /* TTYI */
      {
         if (dev->cbreak) CLR_EQ;
	 else SET_EQ;
	 dev->cbreak = FALSE;
      }
      else if (disp == 12) /* RRQ */
      {
	 CLR_EQ;
	 if (!dev->intenabled)
	 {
	    int c;
	    if (dev->fd == stdin)
	    {
	       c = sim_os_poll_kbd();
	       if (c == 0x03)
	       {
		  dev->cbreak = TRUE;
		  c = EOF;
	       }
	    }
	    else
	       c = fgetc (dev->fd);
	    if (c != EOF)
	    {
	       dev->inchar = c;
	       SET_EQ;
	    }
	 }
	 else if (dev->select)
	 {
	    SET_EQ;
	 }
      }
      break;

   case PRINTER:
#ifdef DEBUGPRINTER
      fprintf (stderr, "CRU %04X TB %d\n", devaddr, disp);
#endif
   case CONOUT:
      switch (disp)
      {
      case 11: /* WRQ */
      case 13: /* DCD */
      case 14: /* DSR */
	 SET_EQ;
	 break;
      default:
	 CLR_EQ;
      }
      break;

   case SERIAL:
      switch (disp)
      {
      case 0: /* TTYI */
         if (dev->cbreak) CLR_EQ;
	 else SET_EQ;
	 dev->cbreak = FALSE;
	 break;
      case 12: /* RRQ */
	 CLR_EQ;
	 if (!dev->intenabled)
	 {
	    int c;
	    c = fgetc (dev->fd);
	    if (c != EOF)
	    {
	       dev->inchar = c;
	       SET_EQ;
	    }
	 }
	 else if (dev->select)
	 {
	    SET_EQ;
	 }
	 break;
      case 11: /* WRQ */
      case 13: /* DCD */
      case 14: /* DSR */
	 SET_EQ;
	 break;
      default:
	 CLR_EQ;
      }
      break;

   case FLOPPY:
#ifdef DEBUGFLOPPY
      fprintf (stderr, "CRU %04X TB %d\n", devaddr, disp);
#endif
      if (devaddr == dev->devaddr) /* Data Transfer */
      {
	 switch (disp)
	 {
	 case 16: /* OP complete */
	 case 17: /* Xfer ready (XFER) */
	    SET_EQ;
	    break;
	 case 26: /* Controller busy (CTLBSY) */
	 default:
	    CLR_EQ;
	 }
      }
      else /* Command */
      {
	 switch (disp)
	 {
	 case 0: /* OP complete (OPCOMP) */
	 case 1: /* Xfer ready (CXFER) */
	 case 15: /* Controller busy (CBUSY) */
	    SET_EQ;
	    break;
	 case 10: /* Controller busy (CBUSY) */
	 default:
	    CLR_EQ;
	 }
      }
      break;

   case CARDREADER:
#ifdef DEBUGCARD
      fprintf (stderr, "CRU %04X TB %d\n", devaddr, disp);
#endif
      switch (disp)
      {
	 case 15: /* Column Ready */
	    SET_EQ;
	    break;
         default: ;
	    CLR_EQ;
      }
      break;

   default: ;
   }
   return;
}

/***********************************************************************
* loadcru - Load CRU.
***********************************************************************/

void
loadcru (uint16 inst, uint16 dat)
{
   Device *dev;
   int i;
   uint16 reg;
   uint16 tag;
   uint16 cnt;
   uint16 devaddr;
   uint8 ch;
   uint8 ch1;

   reg = inst & 0x000F;
   tag = (inst & 0x0030) >> 4;
   cnt = (inst & 0x03C0) >> 6;

   devaddr = R12;

   /*
   ** Front panel is always attached
   */

   if (devaddr == PANEL)
   {
      ch = dat;
      lights = (lights << 8) | ch;
      return;
   }
   
#ifdef DEBUGCRU
   fprintf (stderr, "LDCR %04X CNT %d DAT %04X\n", devaddr, cnt, dat);
#endif

   if ((i = finddev (devaddr)) < 0) return;
   if (devices[i].type == CONIN)
   {
      if ((i = findtype (CONOUT, "CO")) < 0) return;
   }
   dev = &devices[i];

   switch (dev->type)
   {
   case SERIAL:
   case PRINTER:
#ifdef DEBUGPRINTER
      fprintf (stderr, "LDCR %04X CNT %d", devaddr, cnt);
      if (cnt > 1 && cnt < 9)
      {
	 ch = dat & 0xFF;
	 if (!isprint(ch)) ch = '.';
	 fprintf (stderr, " DATA %04X %c\n", dat, ch);
      }
      else
      {
	 ch = (dat >> 8) & 0xFF;
	 if (!isprint(ch)) ch = '.';
	 ch1 = dat & 0xFF;
	 if (!isprint(ch1)) ch1 = '.';
	 fprintf (stderr, " DATA %04X %c%c\n", dat, ch, ch1);
      }
#endif
      ch = dat & 0xFF;
      if (isprint(ch) || (ch > 0x07 && ch < 0x0D)) fputc (ch, dev->fd);
      if (dev->intlvl > 0)
	 intrequest = intrequest | (1 << dev->intlvl);
      break;

   case CONOUT:
      ch = dat;
      if (dev->fd == stdout)
	 sim_os_putchar (ch);
      else
         fputc (ch, dev->fd);
      intrequest = intrequest | (1 << dev->intlvl);
      break;

   case FLOPPY:
#ifdef DEBUGFLOPPY
      fprintf (stderr, "LDCR %04X", devaddr);
#endif
      if (devaddr == dev->devaddr) /* Data transfer */
      {
	 for (i = 0; i < devcnt; i++) /* find selected unit */
	    if (devices[i].select && devices[i].devaddr == devaddr)
	       break;
	 if (i == devcnt)
	 {
	    dev->laststat = 0x4;	/* Drive not ready */
	    return;
	 }
	 dev = &devices[i];
#ifdef DEBUGFLOPPY
	 ch = (dat >> 8) & 0xFF;
	 if (!isprint(ch)) ch = '.';
	 ch1 = dat & 0xFF;
	 if (!isprint(ch1)) ch1 = '.';
	 fprintf (stderr, " DATA %04X %c%c\n", dat, ch, ch1);
#endif
	 fputc ((dat >> 8) & 0xFF, dev->fd);
	 fputc (dat & 0xFF, dev->fd);
	 dev->count -= 2;
	 if (dev->count)
	    intrequest = intrequest | (1 << dev->intlvl);
      }
      else /* Command */
      {
	 int diskloc;
         uint16 cmd;
         uint16 unit;
	 uint16 sec;
	 uint16 trk;

         cmd = dat;
#ifdef DEBUGFLOPPY
	 fprintf (stderr, " CMD %04X", cmd);
#endif

	 /* If Clear Status port, return */
	 if (((cmd & 0xF000) >> 12) == 15)
	 {
#ifdef DEBUGFLOPPY
	    fprintf (stderr, "\n");
#endif
	    for (i = 0; i < devcnt; i++) /* find selected unit */
	       if (devices[i].cdevaddr == devaddr)
	       {
		  dev->cbreak = FALSE;
		  dev->select = FALSE;
		  dev->laststat = 0;
	       }
	    return;
	 }

	 unit = (cmd & 0x0C00) >> 10;
#ifdef DEBUGFLOPPY
	 fprintf (stderr, " UNIT %d", unit);
#endif

	 for (i = 0; i < devcnt; i++) /* find selected unit */
	    if (devices[i].cdevaddr == devaddr && devices[i].unit == unit)
	       break;
	 if (i == devcnt)
	 {
#ifdef DEBUGFLOPPY
	    fprintf (stderr, " not selected\n");
#endif
	    dev->laststat = 0x4 | (unit << 13);	/* Drive not ready */
	    intrequest = intrequest | (1 << dev->intlvl);
	    return;
	 }
	 dev = &devices[i];

	 switch ((cmd & 0xF000) >> 12)
	 {
	 case 0: /* Select */
	    dev->select = TRUE;
	    break;
	 case 1: /* Seek */
	 case 14: /* IPL */
	    trk = cmd & 0x007F;
	    diskloc = (FPYSECTLEN * FPYSECTTRK) * trk;
	    if (fseek (dev->fd, diskloc, SEEK_SET) < 0)
	    {
	       sprintf (view[0], "seek failed: %s",
		     strerror(errno));
	       sprintf (view[1], "filename: %s", dev->file);
	       run = FALSE;
	       return;
	    }
#ifdef DEBUGFLOPPY
	    fprintf (stderr, " TRACK %d LOC %ld", trk, diskloc);
#endif
	    dev->track = diskloc;
	    dev->select = TRUE;
	    break;
	 case 2: /* Restore */
	    dev->select = TRUE;
	    dev->track = 0;
	    break;
	 case 3: /* Sector Length */
	    dev->select = TRUE;
	    break;
	 case 4: /* Read */
	 case 6: /* Read Unformatted */
	 case 7: /* Write */
	 case 8: /* Write Delete */
	    sec = cmd & 0x001F;
	    diskloc = dev->track + (FPYSECTLEN * (sec-1));
#ifdef DEBUGFLOPPY
	    fprintf (stderr, " SECTOR %d LOC %ld", sec, diskloc);
#endif
	    if (fseek (dev->fd, diskloc, SEEK_SET) < 0)
	    {
	       sprintf (view[0], "seek failed: %s",
		     strerror(errno));
	       sprintf (view[1], "filename: %s", dev->file);
	       run = FALSE;
	       return;
	    }
	    dev->select = TRUE;
	    break;
	 case 5: /* Read ID */
	    break;
	 case 9: /* Format Track */
	    trk = cmd & 0x007F;
	    diskloc = (FPYSECTLEN * FPYSECTTRK) * trk;
#ifdef DEBUGFLOPPY
	    fprintf (stderr, " TRACK %d LOC %ld", trk, diskloc);
#endif
	    if (fseek (dev->fd, diskloc, SEEK_SET) < 0)
	    {
	       sprintf (view[0], "seek failed: %s",
		     strerror(errno));
	       sprintf (view[1], "filename: %s", dev->file);
	       run = FALSE;
	       return;
	    }
	    for (i = 0; i < FPYSECTTRK*FPYSECTLEN; i++) fputc ('\0', dev->fd);
	    dev->track = trk+1;
	    dev->select = TRUE;
	    break;
	 case 10: /* Load Int Mask */
	    dev->select = TRUE;
	    break;
	 case 11: /* Stop */
	    dev->select = TRUE;
	    dev->cbreak = TRUE;
	    dev->count = 0;
	    break;
	 case 12: /* Step Head */
	    break;
	 default: ; /* Maint. commands */
	 }
#ifdef DEBUGFLOPPY
	 fprintf (stderr, "\n");
#endif
	 intrequest = intrequest | (1 << dev->intlvl);
      }
      break;

   case CARDREADER:
#ifdef DEBUGCARD
      fprintf (stderr, "LDCR %04X CNT %d DATA %04X\n", devaddr, cnt, dat);
#endif
      break;

   default: ;
   }

   return;
}

/***********************************************************************
* storecru - Store CRU.
***********************************************************************/

uint16
storecru (uint16 inst)
{
   Device *dev;
   int i;
   uint16 reg;
   uint16 tag;
   uint16 cnt;
   uint16 devaddr;
   uint16 dat;
   uint8 ch;
   uint8 ch1;

   reg = inst & 0x000F;
   tag = (inst & 0x0030) >> 4;
   cnt = (inst & 0x03C0) >> 6;

   dat = 0;
   devaddr = R12;


   if ((i = finddev (devaddr)) < 0) return(-1);
   if (devices[i].type == CONOUT) 
   {
      if ((i = findtype (CONIN, "CI")) < 0) return;
   }
   dev = &devices[i];

   switch (dev->type)
   {
   case PRINTER:
#ifdef DEBUGPRINTER
      fprintf (stderr, "STCR %04X CNT %d\n", devaddr, cnt);
#endif
      break;

   case SERIAL:
   case CONIN:
      dev->select = 0;
      dat = dev->inchar;
      break;

   case FLOPPY:
#ifdef DEBUGFLOPPY
      fprintf (stderr, "STCR %04X", devaddr);
#endif
      if (devaddr == dev->devaddr) /* Data transfer */
      {
         uint16 d;
	 uint16 c;

	 for (i = 0; i < devcnt; i++) /* find selected unit */
	    if (devices[i].select && devices[i].devaddr == devaddr)
	       break;
	 if (i == devcnt)
	 {
	    dev->laststat = 0x4;	/* Drive not ready */
	    return (dat);
	 }
	 dev = &devices[i];
	 d = fgetc (dev->fd);
	 c = fgetc (dev->fd);
	 d = (d << 8) | c;
	 dat = d;
#ifdef DEBUGFLOPPY
	 ch = (dat >> 8) & 0xFF;
	 if (!isprint(ch)) ch = '.';
	 ch1 = dat & 0xFF;
	 if (!isprint(ch1)) ch1 = '.';
	 fprintf (stderr, " DATA %04X %c%c\n", dat, ch, ch1);
#endif
	 dev->count -= 2;
	 if (dev->count)
	    intrequest = intrequest | (1 << dev->intlvl);
      }
      else /* Command */
      {
	 uint16 unit = 0;

	 for (i = 0; i < devcnt; i++)
	    if (devices[i].select && devices[i].cdevaddr == devaddr)
	    {
	       dev = &devices[i];
	       unit = dev->unit;
#ifdef DEBUGFLOPPY
	       fprintf (stderr, " UNIT %d", unit);
#endif

	       if (dev->cbreak)
	       {
		  dev->cbreak = FALSE;
		  dev->select = FALSE;
	       }
	       break;
	    }
	 if (dev->laststat)
	    dat = dev->laststat | 0x8001;
	 else
	    dat = 0x8003 | (unit << 13);

#ifdef DEBUGFLOPPY
	 fprintf (stderr, " DATA %04X\n", dat);
#endif
      }
      break;

   case CARDREADER:
      {
	 int cdat;
	 if (dev->fd)
	 {
	    do {
	       if ((cdat = fgetc (dev->fd)) < 0)
	       {
		  fclose (dev->fd);
		  dev->fd = NULL;
		  cdat = ' ';
	       }
	    } while (cdat < 0x20 || cdat > 0x7F);
	 }
	 else cdat = ' ';
	 ch = cdat;
	 dat = tocard[cdat];
#ifdef DEBUGCARD
	 fprintf (stderr, "STCR %04X CNT %d DATA %02X(%c) - %04X\n",
		  devaddr, cnt, ch, ch, cdat);
#endif
      }
      break;

   default: ;
   }
#ifdef DEBUGCRU
   fprintf (stderr, "STCR %04X CNT %d DAT %04X\n", devaddr, cnt, dat);
#endif
   return (dat);
}


/***********************************************************************
* startclk - Start clock.
***********************************************************************/

void
startclk (void)
{
#if defined(UNIX)
   pthread_t thread_handle;
#endif

#ifdef DEBUGINTS
   fprintf (stderr, "CKON called\n");
#endif
   /*
   ** First call starts the thread
   */

   if (!clockthreadactive)
   {
      clockthreadactive = TRUE;
#if defined(UNIX)
      if (pthread_create (&thread_handle,
			  NULL,
			  (void *(*)(void *)) clock_thread,
			  NULL) < 0)
#endif
#if defined(_WIN32)
      if (_beginthread (clock_thread,
			NULL,
			NULL) < 0)
#endif
      {
	 perror ("startclk: Can't start clock thread");
      }
   }
   clockactive = TRUE;
}

/***********************************************************************
* stopclk - Stop clock.
***********************************************************************/

void
stopclk (void)
{
#ifdef DEBUGINTS
   fprintf (stderr, "CKOF called\n");
#endif
   clockactive = FALSE;
   intrequest = intrequest & ~(1 << CLOCKINT);
}

/***********************************************************************
* boot - Boot from device.
***********************************************************************/

int
boot (char *device)
{
   Device *dev;
   uint16 lp;
   uint16 val;
   int i;
   int sect;

   run = FALSE;
   lp = LOADADDRESS;

   while (isspace (*device)) device++;
   if (islower(*device)) *device = toupper (*device);

   for (i = 0; i < devcnt; i++)
   {
      if (!strcmp(devices[i].dev, device)) break;
   }

   if (i == devcnt)
   {
      sprintf (view[0], "Boot device %s is not attached", device);
      return (-1);
   }

   dev = &devices[i];

   if (fseek (dev->fd, 0, SEEK_SET) < 0)
   {
      sprintf (view[0], "boot: seek failed: %s",
	    strerror(errno));
      sprintf (view[1], "filename: %s", dev->file);
      return (-1);;
   }

   do {

      for (i = 0; i < FPYSECTLEN; i+=2)
      {
	 int c;

	 if ((c = fgetc (dev->fd)) == EOF)
	 {
	    sprintf (view[0], "boot: read failed: %s",
		  strerror(errno));
	    sprintf (view[1], "filename: %s", dev->file);
	    return (-1);;
	 }
	 val = (c & 0xFF) << 8;
	 if ((c = fgetc (dev->fd)) == EOF)
	 {
	    sprintf (view[0], "boot: read failed: %s",
		  strerror(errno));
	    sprintf (view[1], "filename: %s", dev->file);
	    return (-1);;
	 }
	 val = val | (c & 0xFF);
	 PUTMEM (lp, val);
	 lp += 2;
	 if (lp >= MEMLEN)
	 {
	    sprintf (view[0], "boot: image larger than memory");
	    sprintf (view[1], "filename: %s", dev->file);
	    return (-1);;
	 }
      }

      sect++;
      lp -= 2;

   } while (sect < (FPYSECTTRK * FPYTRKCYL) && (val != 0xFFFF));

#ifdef DUMPBOOT
   HEXDUMP (stderr, &memory[LOADADDRESS], lp - LOADADDRESS, 0xF1F0);
#endif

   run = TRUE;
   wpreg = 0x80;
   PUTMEM (LOADADDRESS, (dev->unit << 10));
   pcreg = LOADADDRESS + 2;

   return (0);
}

/***********************************************************************
* checkinterrupts - If any pending and not masked, take it.
***********************************************************************/

void 
checkinterrupts (void)
{
   int curlvl;
   int i;

#ifdef DEBUGINTS
   fprintf (stderr, "checkinterrupts called: idle = %s, lvl = %d\n",
	    idle ? "TRUE" : "FALSE", statreg & 0xF);
#endif
#if 1
   do
   {
#endif
      if (intrequest)
      {
	 curlvl = (statreg & 0xF) + 1;
	 for (i = 0; i < curlvl; i++)
	 {
	    if (intrequest & (1 << i))
	    {
	       uint16 opc;
	       uint16 owp;
	       uint16 ost;

	       if (GETMEM(pcreg) == 0x340) pcreg += 2;
	       ost = statreg;
	       opc = pcreg;
	       owp = wpreg;
	       SET_MASK(i-1);
	       intrequest = intrequest & ~(1 << i);
	       wpreg = GETMEM (i * 4) & 0xFFFE;
	       pcreg = GETMEM (i * 4 + 2) & 0xFFFE;
	       PUTREG (13, owp);
	       PUTREG (14, opc);
	       PUTREG (15, ost);
	       if (model >= 10)
	       {
		  CLR_NONPRV;
	       }
	       idle = FALSE;
	       return;
	    }
	 }
      }
      if ((i = findtype (CONIN, "CI")) >= 0)
      {
	 Device *dev;
         int c;

	 dev = &devices[i];
	 if (dev->intenabled)
	 {
	    c = sim_os_poll_kbd();
	    if (c != EOF)
	    {
	       if (c == 0x03) dev->cbreak = TRUE;
	       else 
	       {
		  dev->select = 1;
		  dev->inchar = c;
		  intrequest = intrequest | (1 << dev->intlvl);
	       }
	    }
	 }
      }
#if 1
   } while (idle);
#endif
}
