/***********************************************************************
*
* simops.c - Simulate instruction operations for the TI 990 computer.
*
* Changes:
*   05/29/03   DGP   Original.
*   06/22/03   DGP   Fixed XOP REG function.
*   07/08/03   DGP   Fixed status on Subtract and indexing with negative
*                    values.
*   07/09/03   DGP   Forced PC and WP to be on even boundry.
*   07/13/03   DGP   Forced word and address access to even boundry.
*
***********************************************************************/

#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>

#include "simdef.h"

extern uint16 pcreg;	/* The program PC */
extern uint16 statreg;	/* The program status register */
extern uint16 wpreg;	/* The program Workspace Pointer */
extern uint16 lights;	/* The panel lights */
extern uint16 breakaddr; /* The break point address */
extern long instcount;
extern long pancount;	/* Instructions executed for panel */

extern int run;
extern int idle;
extern int model;
extern int breakflag;
extern int enablepanel;
extern int traceflag;
extern int deferint;
extern char view[MAXVIEW][81];
extern FILE *tracefd;

extern uint8 memory[MEMSIZE];

static void runinst (uint16);

/***********************************************************************
* decodeinst - Decode instructions.
***********************************************************************/

static int
decodeinst (uint16 inst)
{
   if (inst == 0x0300) return TYPE_8;
   if (inst & 0xC000) return TYPE_1;
   if ((inst & 0xF800) == 0x3000) return TYPE_4;
   if ((inst & 0xF000) == 0x2000) return TYPE_3;
   if ((inst & 0xF000) == 0x3000) return TYPE_9;
   if ((inst & 0xF000) == 0x1000) return TYPE_2;
   if ((inst & 0xFC00) == 0x0800) return TYPE_5;
   if ((inst & 0xFC00) == 0x0400) return TYPE_6;
   if ((inst & 0xFF00) == 0x0200) return TYPE_8;
   if ((inst & 0xFF00) == 0x0200) return TYPE_8;
   if ((inst & 0xFF00) == 0x0300) return TYPE_7;
   if ((inst & 0xFF00) == 0x0000) return TYPE_7;
   if ((inst & 0xFFF0) == 0x0C00) return TYPE_7;
   return (TYPE_ILL);
}

/***********************************************************************
* putmem - Put a word in memory.
***********************************************************************/

void
putmem (uint16 m, uint16 v)
{
   if (m < MEMLEN)
   {
      m &= 0xFFFE;
      memory[(m)+MSB] = (v) >> 8 & 0xFF;
      memory[(m)+LSB] = (v) & 0xFF;
   }
}

/***********************************************************************
* putbyte - Put a byte in memory.
***********************************************************************/

void
putmemb (uint16 m, uint8 v)
{
   if (m < MEMLEN)
      memory[(m)] = (v) & 0xFF; 
}

/***********************************************************************
* getaddr - Get an effective address.
***********************************************************************/

static uint16
getaddr (uint16 sr, uint16 st)
{
   uint16 sa = 0;

   switch (st)
   {
   case 0:
      if (traceflag)
	 fprintf (tracefd, " R%d", sr);
      sa = wpreg + (sr * 2);
      break;
   case 1:
      if (traceflag)
	 fprintf (tracefd, " *R%d", sr);
      sa = GETREG (sr);
      break;
   case 2:
      sa = GETINST;
      if (traceflag)
	 fprintf (tracefd, " @>%X(%d)", sa, sr);
      pcreg += 2;
      if (sr != 0) (int16)sa += (int16)GETREG(sr);
      break;
   case 3:
      if (traceflag)
	 fprintf (tracefd, " *R%d+", sr);
      sa = GETREG (sr);
      PUTREG (sr, GETREG (sr) + 2);
      break;
   }
   return (sa & 0xFFFE);
}

/***********************************************************************
* getword - Get a word operand. If memory access is past the end of
* defined memory, return -1;
***********************************************************************/

uint16
getword (uint16 sr, uint16 st, int inc)
{
   uint16 sa;
   uint16 sval;

   switch (st)
   {
   case 0:
      if (traceflag && inc)
	 fprintf (tracefd, " R%d", sr);
      sval = GETREG (sr);
      break;
   case 1:
      if (traceflag && inc)
	 fprintf (tracefd, " *R%d", sr);
      sa = GETREG (sr);
      if (sa >= MEMSIZE) sval = -1;
      else sval = GETMEM (sa & 0xFFFE);
      break;
   case 2:
      sa = GETINST;
      if (traceflag && inc)
	 fprintf (tracefd, " @>%X(%d)", sa, sr);
      if (inc) pcreg += 2;
      if (sr != 0) (int16)sa += (int16)GETREG(sr);
      if (sa >= MEMSIZE) sval = -1;
      else sval = GETMEM (sa & 0xFFFE);
      break;
   case 3:
      if (traceflag && inc)
	 fprintf (tracefd, " *R%d+", sr);
      sa = GETREG (sr);
      if (sa >= MEMSIZE) sval = -1;
      else sval = GETMEM (sa & 0xFFFE);
      if (inc) PUTREG (sr, GETREG (sr) + 2);
      break;
   }
   return (sval);
}

/***********************************************************************
* getbyte - Get a byte operand. If memory access is past the end of
* defined memory, return -1;
***********************************************************************/

uint8
getbyte (uint16 sr, uint16 st, int inc)
{
   uint16 sa;
   uint8 sval;

   switch (st)
   {
   case 0:
      if (traceflag && inc)
	 fprintf (tracefd, " R%d", sr);
      sval = (GETREG (sr) & 0xFF00) >> 8;
      break;
   case 1:
      if (traceflag && inc)
	 fprintf (tracefd, " *R%d", sr);
      sa = GETREG (sr);
      if (sa >= MEMSIZE) sval = -1;
      else sval = GETMEMB (sa);
      break;
   case 2:
      sa = GETINST;
      if (traceflag && inc)
	 fprintf (tracefd, " @>%X(%d)", sa, sr);
      if (inc) pcreg += 2;
      if (sr != 0) (int16)sa += (int16)GETREG(sr);
      if (sa >= MEMSIZE) sval = -1;
      else sval = GETMEMB (sa);
      break;
   case 3:
      if (traceflag && inc)
	 fprintf (tracefd, " *R%d+", sr);
      sa = GETREG (sr);
      if (sa >= MEMSIZE) sval = -1;
      else sval = GETMEMB (sa);
      if (inc) PUTREG (sr, GETREG (sr) + 1);
      break;
   }
   return (sval);
}

/***********************************************************************
* putword - Put a word operand.  If memory access is beyond allowed 
* memory (in ROM space) it is ignored.
***********************************************************************/

void
putword (uint16 dr, uint16 dt, uint16 dval)
{
   uint16 da;

   switch (dt)
   {
   case 0:
      if (traceflag)
	 fprintf (tracefd, " R%d", dr);
      PUTREG (dr, dval);
      break;
   case 1:
      if (traceflag)
	 fprintf (tracefd, " *R%d", dr);
      da = GETREG(dr);
      if (da < MEMLEN)
	 PUTMEM (da, dval);
      break;
   case 2:
      da = GETINST;
      if (traceflag)
	 fprintf (tracefd, " @>%X(%d)", da, dr);
      pcreg += 2;
      if (dr != 0) (int16)da += (int16)GETREG (dr);
      if (da < MEMLEN)
	 PUTMEM (da, dval);
      break;
   case 3:
      if (traceflag)
	 fprintf (tracefd, " *R%d+", dr);
      da = GETREG(dr);
      if (da < MEMLEN)
	 PUTMEM (da, dval);
      PUTREG (dr, GETREG (dr) + 2);
      break;
   }
}

/***********************************************************************
* putbyte - Put a byte operand.  If memory access is beyond allowed 
* memory (in ROM space) it is ignored.
***********************************************************************/

void
putbyte (uint16 dr, uint16 dt, uint8 dval)
{
   uint16 da;

   switch (dt)
   {
   case 0:
      if (traceflag)
	 fprintf (tracefd, " R%d", dr);
      da = (GETREG (dr) & 0x00FF) | (dval << 8);
      PUTREG (dr, da);
      break;
   case 1:
      if (traceflag)
	 fprintf (tracefd, " *R%d", dr);
      da = GETREG(dr);
      if (da < MEMLEN)
	 PUTMEMB (da, dval);
      break;
   case 2:
      da = GETINST;
      if (traceflag)
	 fprintf (tracefd, " @>%X(%d)", da, dr);
      pcreg += 2;
      if (dr != 0) (int16)da += (int16)GETREG (dr);
      if (da < MEMLEN)
	 PUTMEMB (da, dval);
      break;
   case 3:
      if (traceflag)
	 fprintf (tracefd, " *R%d+", dr);
      da = GETREG(dr);
      if (da < MEMLEN)
	 PUTMEMB (da, dval);
      PUTREG (dr, GETREG (dr) + 1);
      break;
   }
}

/***********************************************************************
* chkparity - Check parity of a byte.
***********************************************************************/

static void
chkparity (uint8 bval)
{
   int c;
   int i;

   for (c = 0, i = 0; i < 8; i++)
   {
      c = c + (bval & 0x01);
      bval = bval >> 1;
   }
   if (c & 0x01) SET_ODDP;
   else CLR_ODDP;
}

/***********************************************************************
* cmpzero - Compare value to zero.
***********************************************************************/

static void
cmpzero (uint16 val)
{
   if (val > 0) SET_LGT;
   else CLR_LGT;
   if ((int16)val > 0) SET_AGT;
   else CLR_AGT;
   if (val == 0) SET_EQ;
   else CLR_EQ;
}

/***********************************************************************
* cmpbzero - Compare byte value to zero.
***********************************************************************/

static void
cmpbzero (uint8 bval)
{
   if (bval > 0) SET_LGT;
   else CLR_LGT;
   if ((int8)bval > 0) SET_AGT;
   else CLR_AGT;
   if (bval == 0) SET_EQ;
   else CLR_EQ;
}


/***********************************************************************
* proc1op - Process type 1 operations.
***********************************************************************/

static int
proc1op (uint16 inst)
{
   int32 acc;
   uint16 st;
   uint16 sr;
   uint16 val;
   uint16 cval;
   uint16 dval;
   uint16 dt;
   uint16 dr;
   uint16 op;
   int16 sv1, sv2;
   uint8 bval;
   uint8 bcval;
   int8 bsv1, bsv2;

   sr =  inst & 0x000F;
   st = (inst & 0x0030) >> 4;
   dr = (inst & 0x03C0) >> 6;
   dt = (inst & 0x0C00) >> 10;
   op = (inst & 0xF000) >> 12;

   switch ((inst & 0xF000) >> 12)
   {
   case 4: /* SZC */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X SZC",
	       pcreg-2, wpreg, statreg, inst);
      val = ~getword (sr, st, TRUE);
      val = getword (dr, dt, FALSE) & val;
      putword (dr, dt, val);
      cmpzero (val);
      break;

   case 5: /* SZCB */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X SZCB",
	       pcreg-2, wpreg, statreg, inst);
      bval = ~getbyte (sr, st, TRUE);
      bval = getbyte (dr, dt, FALSE) & bval;
      putbyte (dr, dt, bval);
      cmpbzero (bval);
      chkparity (bval);
      break;

   case 6: /* S */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X S",
	       pcreg-2, wpreg, statreg, inst);
      val = getword (sr, st, TRUE);
      acc = val;
      cval = getword (dr, dt, FALSE);
      acc = cval - acc;
      if (acc & 0xFFFF0000) SET_CARRY;
      else CLR_CARRY;
      if (((val & 0x8000) == (cval & 0x8000)) && (acc & 0x8000 != val &0x8000))
         SET_OVER;
      else CLR_OVER;
      dval = acc & 0xFFFF;
      putword (dr, dt, dval);
      cmpzero (dval);
      break;

   case 7: /* SB */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X SB",
	       pcreg-2, wpreg, statreg, inst);
      bcval = getbyte (sr, st, TRUE);
      acc = bcval;
      bval = getbyte (dr, dt, FALSE);
      acc = bval - acc;
      if (acc & 0xFFFFFF00) SET_CARRY;
      else CLR_CARRY;
      if (((bval & 0x80) == (bcval & 0x80)) && (acc & 0x80 != bval &0x80))
         SET_OVER;
      else CLR_OVER;
      bval = acc &0xFF;
      putbyte (dr, dt, bval);
      cmpbzero (bval);
      chkparity (bval);
      break;

   case 8: /* C */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X C",
	       pcreg-2, wpreg, statreg, inst);
      val = getword (sr, st, TRUE);
      cval = getword (dr, dt, TRUE);
      if (val > cval) SET_LGT;
      else CLR_LGT;
      if ((int16)val > (int16)cval) SET_AGT;
      else CLR_AGT;
      if (val == cval) SET_EQ;
      else CLR_EQ;
      break;

   case 9: /* CB */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X CB",
	       pcreg-2, wpreg, statreg, inst);
      bval = getbyte (sr, st, TRUE);
      bcval = getbyte (dr, dt, TRUE);
      if (bval > bcval) SET_LGT;
      else CLR_LGT;
      if ((int8)bval > (int8)bcval) SET_AGT;
      else CLR_AGT;
      if (bval == bcval) SET_EQ;
      else CLR_EQ;
      chkparity (bval);
      break;

   case 10: /* A */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X A",
	       pcreg-2, wpreg, statreg, inst);
      sv1 = (int16)getword (sr, st, TRUE);
      acc = sv1;
      sv2 = (int16)getword (dr, dt, FALSE);
      acc = acc + sv2;
      if (acc & 0xFFFF0000) SET_CARRY;
      else CLR_CARRY;
      if (((val & 0x8000) == (cval & 0x8000)) && (acc & 0x8000 != val &0x8000))
         SET_OVER;
      else CLR_OVER;
      val = acc & 0xFFFF;
      putword (dr, dt, val);
      cmpzero (val);
      break;

   case 11: /* AB */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X AB",
	       pcreg-2, wpreg, statreg, inst);
      bsv1 = (int8)getbyte (sr, st, TRUE);
      acc = bsv1;
      bsv2 = (int8)getbyte (dr, dt, FALSE);
      acc = acc + bsv2;
      if (acc & 0xFFFFFF00) SET_CARRY;
      else CLR_CARRY;
      if (((bval & 0x80) == (bcval & 0x80)) && (acc & 0x80 != bval &0x80))
         SET_OVER;
      else CLR_OVER;
      bval = acc & 0xFF;
      putbyte (dr, dt, bval);
      cmpbzero (bval);
      chkparity (bval);
      break;

   case 12: /* MOV */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X MOV",
	       pcreg-2, wpreg, statreg, inst);
      val = getword (sr, st, TRUE);
      putword (dr, dt, val);
      cmpzero (val);
      break;

   case 13: /* MOVB */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X MOVB",
	       pcreg-2, wpreg, statreg, inst);
      bval = getbyte (sr, st, TRUE);
      putbyte (dr, dt, bval);
      cmpbzero (bval);
      chkparity (bval);
      break;

   case 14: /* SOC */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X SOC",
	       pcreg-2, wpreg, statreg, inst);
      val = getword (sr, st, TRUE);
      val = getword (dr, dt, FALSE) | val;
      putword (dr, dt, val);
      cmpzero (val);
      break;

   case 15: /* SOCB */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X SOCB",
	       pcreg-2, wpreg, statreg, inst);
      bval = getbyte (sr, st, TRUE);
      bval = getbyte (dr, dt, FALSE) | bval;
      putbyte (dr, dt, bval);
      cmpbzero (bval);
      chkparity (bval);
      break;

   default:
      return (-1);
   }
   return (0);
}

/***********************************************************************
* proc2op - Process type 2 operations.
***********************************************************************/

static int
proc2op (uint16 inst)
{
   int8 disp;
   int16 wpc;

   disp = inst & 0x00FF;
   wpc = (pcreg >> 1) & 0x7FFF;

   switch ((inst & 0x0F00) >> 8)
   {
   case 0: /* JMP */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X JMP %X",
	       pcreg-2, wpreg, statreg, inst, disp & 0xFF);
      break;

   case 1: /* JLT */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X JLT %X",
	       pcreg-2, wpreg, statreg, inst, disp & 0xFF);
      if (!IS_AGT && !IS_EQ) break;
      return (0);

   case 2: /* JLE */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X JLE %X",
	       pcreg-2, wpreg, statreg, inst, disp & 0xFF);
      if (!IS_LGT || IS_EQ) break;
      return (0);

   case 3: /* JEQ */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X JEQ %X",
	       pcreg-2, wpreg, statreg, inst, disp & 0xFF);
      if (IS_EQ) break;
      return (0);

   case 4: /* JHE */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X JHE %X",
	       pcreg-2, wpreg, statreg, inst, disp & 0xFF);
      if (IS_LGT || IS_EQ) break;
      return (0);

   case 5: /* JGT */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X JGT %X",
	       pcreg-2, wpreg, statreg, inst, disp & 0xFF);
      if (IS_AGT) break;
      return (0);

   case 6: /* JNE */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X JNE %X",
	       pcreg-2, wpreg, statreg, inst, disp & 0xFF);
      if (!IS_EQ) break;
      return (0);

   case 7: /* JNC */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X JNC %X",
	       pcreg-2, wpreg, statreg, inst, disp & 0xFF);
      if (!IS_CARRY) break;
      return (0);

   case 8: /* JOC */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X JOC %X",
	       pcreg-2, wpreg, statreg, inst, disp & 0xFF);
      if (IS_CARRY) break;
      return (0);

   case 9: /* JNO */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X JNO %X",
	       pcreg-2, wpreg, statreg, inst, disp & 0xFF);
      if (!IS_OVER) break;
      return (0);

   case 10: /* JL */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X JL %X",
	       pcreg-2, wpreg, statreg, inst, disp & 0xFF);
      if (!IS_LGT && !IS_EQ) break;
      return (0);

   case 11: /* JH */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X JH %X",
	       pcreg-2, wpreg, statreg, inst, disp & 0xFF);
      if (IS_LGT && !IS_EQ) break;
      return (0);

   case 12: /* JOP */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X JOP %X",
	       pcreg-2, wpreg, statreg, inst, disp & 0xFF);
      if (IS_ODDP) break;
      return (0);

   case 13: /* SBO */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X SBO %X",
	       pcreg-2, wpreg, statreg, inst, disp & 0xFF);
      if (model >= 10 && IS_NONPRV)
      {
      }
      setbitone (inst);
      return (0);

   case 14: /* SBZ */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X SBZ %X",
	       pcreg-2, wpreg, statreg, inst, disp & 0xFF);
      if (model >= 10 && IS_NONPRV)
      {
      }
      setbitzero (inst);
      return (0);

   case 15: /* TB */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X TB %X",
	       pcreg-2, wpreg, statreg, inst, disp & 0xFF);
      if (model >= 10 && IS_NONPRV)
      {
      }
      testbit (inst);
      return (0);

   default:
      return (-1);
   }

   /*
   ** Fall through, take the jump
   */

   if ((uint8)disp == 0xFF)
   {
      idle = TRUE;
      if ((statreg & 0x000F) == 0) run = FALSE;
   }

   wpc += disp;
   pcreg = (wpc << 1) & 0xFFFE;
   return (0);
}

/***********************************************************************
* proc3op - Process type 3 operations.
***********************************************************************/

static int
proc3op (uint16 inst)
{
   uint16 st;
   uint16 sr;
   uint16 reg;
   uint16 val;
   uint16 cval;
   uint16 op;

   sr =  inst & 0x000F;
   st = (inst & 0x0030) >> 4;
   reg = (inst & 0x03C0) >> 6;
   op = (inst & 0x1C00) >> 10;

   switch (op)
   {
   case 0: /* COC */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X COC",
	       pcreg-2, wpreg, statreg, inst);
      val = getword (sr, st, TRUE);
      if (traceflag)
         fprintf (tracefd, " R%d", reg);
      cval = GETREG (reg) & val;
      if (cval == val) SET_EQ;
      else CLR_EQ;
      break;

   case 1: /* CZC */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X CZC",
	       pcreg-2, wpreg, statreg, inst);
      val = getword (sr, st, TRUE);
      if (traceflag)
         fprintf (tracefd, " R%d", reg);
      cval = ~GETREG (reg) & val;
      if (cval == val) SET_EQ;
      else CLR_EQ;
      break;

   case 2: /* XOR */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X XOR",
	       pcreg-2, wpreg, statreg, inst);
      val = getword (sr, st, TRUE);
      if (traceflag)
         fprintf (tracefd, " R%d", reg);
      val = GETREG (reg) ^ val;
      PUTREG (reg, val);
      cmpzero (val);
      break;

   case 3: /* XOP */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X XOP",
	       pcreg-2, wpreg, statreg, inst);
      val = wpreg;
      cval = getaddr (sr, st);
      if (traceflag)
         fprintf (tracefd, " %d", reg);
      wpreg = GETMEM (reg * 4 + 0x0040) & 0xFFFE;
      PUTREG (11, cval);
      cval = pcreg;
      pcreg = GETMEM (reg * 4 + 0x0042) & 0xFFFE;
      PUTREG (13, val);
      PUTREG (14, cval);
      PUTREG (15, statreg);
      SET_XOP;
      if (model >= 10)
      {
	 CLR_NONPRV;
      }
      deferint = 2;
      break;

   default:
      return (-1);
   }
   return (0);
}

/***********************************************************************
* proc4op - Process type 4 operations.
***********************************************************************/

static int
proc4op (uint16 inst)
{
   uint16 reg;
   uint16 tag;
   uint16 cnt;
   uint16 dat;
   uint8 bdat;

   reg = inst & 0x000F;
   tag = (inst & 0x0030) >> 4;
   cnt = (inst & 0x03C0) >> 6;

   switch ((inst & 0x0400) >> 10)
   {
   case 0: /* LDCR */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X LDCR",
	       pcreg-2, wpreg, statreg, inst);
      if (cnt > 1 && cnt < 9)
      {
         bdat = getbyte (reg, tag, TRUE);
	 cmpbzero (bdat);
	 chkparity(bdat);
	 dat = bdat;
      }
      else
      {
         dat = getword (reg, tag, TRUE);
	 cmpzero (dat);
      }
      loadcru (inst, dat);
      break;

   case 1: /* STCR */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X STCR",
	       pcreg-2, wpreg, statreg, inst);
      dat = storecru (inst);
      if (cnt > 1 && cnt < 9)
      {
         bdat = dat;
	 cmpbzero (bdat);
	 chkparity(bdat);
	 putbyte (reg, tag, bdat);
      }
      else
      {
	 cmpzero (dat);
	 putword (reg, tag, dat);
      }
      break;

   default:
      return (-1);
   }
   return (0);
}

/***********************************************************************
* proc5op - Process type 5 operations.
***********************************************************************/

static int
proc5op (uint16 inst)
{
   uint32 acc;
   uint16 reg;
   uint16 cnt;
   uint16 val;

   reg = inst & 0x000F;
   cnt = (inst & 0x00F0) >> 4;
   if (cnt == 0)
   {
      cnt = R0 & 0x000F;
      if (cnt == 0) cnt = 16;
   }
   val = GETREG (reg);

   switch ((inst & 0x0300) >> 8)
   {
   case 0: /* SRA */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X SRA R%d %d",
	       pcreg-2, wpreg, statreg, inst, reg, cnt);
      acc = val << 16;
      (int32)acc = (int32)acc >> cnt;
      if (acc & 0x00008000) SET_CARRY;
      else CLR_CARRY;
      val = (acc >> 16) & 0xFFFF;
      PUTREG (reg, val);
      cmpzero (val);
      break;

   case 1: /* SRL */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X SRL R%d %d",
	       pcreg-2, wpreg, statreg, inst, reg, cnt);
      acc = val << 16;
      acc = acc >> cnt;
      if (acc & 0x00008000) SET_CARRY;
      else CLR_CARRY;
      val = (acc >> 16) & 0xFFFF;
      PUTREG (reg, val);
      cmpzero (val);
      break;

   case 2: /* SLA */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X SLA R%d %d",
	       pcreg-2, wpreg, statreg, inst, reg, cnt);
      acc = val;
      acc = acc << cnt;
      if ((val & 0x8000) != (acc & 0x8000)) SET_OVER;
      else CLR_OVER;
      if (acc & 0x10000) SET_CARRY;
      else CLR_CARRY;
      val = acc & 0xFFFF;
      PUTREG (reg, val);
      cmpzero (val);
      break;

   case 3: /* SRC */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X SRC R%d %d",
	       pcreg-2, wpreg, statreg, inst, reg, cnt);
      while (cnt)
      {
	 uint16 ob;

         ob = val & 0x0001;
	 val = val >> 1;
	 if (ob)
	 {
	    val = val | 0x8000;
	    SET_CARRY;
	 }
	 else CLR_CARRY;
	 cnt--;
      }
      PUTREG (reg, val);
      cmpzero (val);
      break;

   default:
      return (-1);
   }
   return (0);
}

/***********************************************************************
* proc6op - Process type 6 operations.
***********************************************************************/

static int
proc6op (uint16 inst)
{
   uint16 sr;
   uint16 st;
   uint16 val;
   uint16 sval;

   sr = inst & 0x000F;
   st = (inst & 0x0030) >> 4;

   switch ((inst & 0x03C0) >> 6)
   {
   case 0: /* BLWP */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X BLWP",
	       pcreg-2, wpreg, statreg, inst);
      val = wpreg;
      if (st == 0)
      {
	 if (traceflag)
	    fprintf (tracefd, " R%d", sr);
	 sval = GETREG(sr+1);
         wpreg = GETREG(sr) & 0xFFFE;
      }
      else
      {
	 sval = getaddr (sr, st);
	 wpreg = GETMEM (sval) & 0xFFFE;
	 sval = GETMEM (sval+2);
      }
      PUTREG (13, val);
      PUTREG (14, pcreg);
      PUTREG (15, statreg);
      pcreg = sval & 0xFFFE;
      deferint = 2;
      break;

   case 1: /* B */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X B",
	       pcreg-2, wpreg, statreg, inst);
      if (st == 0)
      {
	 if (traceflag)
	    fprintf (tracefd, " R%d", sr);
	 val = GETREG (sr);
      }
      else val = getaddr (sr, st);
      pcreg = val & 0xFFFE;
      break;

   case 2: /* X */
      {
	 uint16 xinst;

	 if (traceflag)
	    fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X X",
		  pcreg-2, wpreg, statreg, inst);
	 val = getaddr (sr, st);
	 xinst = GETMEM(val);
	 runinst (xinst);
	 deferint = 2;
	 pcreg &= 0xFFFE;
      }
      break;

   case 3: /* CLR */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X CLR",
	       pcreg-2, wpreg, statreg, inst);
      putword (sr, st, 0);
      break;

   case 4: /* NEG */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X NEG",
	       pcreg-2, wpreg, statreg, inst);
      val = getword (sr, st, FALSE);
      if (val == 0x8000) SET_OVER;
      else CLR_OVER;
      val = ~val + 1;
      goto SAVE_CHECK_ZERO;

   case 5: /* INV */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X INV",
	       pcreg-2, wpreg, statreg, inst);
      val = ~getword (sr, st, FALSE);
      goto SAVE_CHECK_ZERO;

   case 6: /* INC */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X INC",
	       pcreg-2, wpreg, statreg, inst);
      sval = getword (sr, st, FALSE);
      val = sval + 1;
      goto CHECK_STATUS;

   case 7: /* INCT */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X INCT",
	       pcreg-2, wpreg, statreg, inst);
      sval = getword (sr, st, FALSE);
      val = sval + 2;
      goto CHECK_STATUS;

   case 8: /* DEC */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X DEC",
	       pcreg-2, wpreg, statreg, inst);
      sval = getword (sr, st, FALSE);
      val = sval - 1;
      goto CHECK_STATUS;

   case 9: /* DECT */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X DECT",
	       pcreg-2, wpreg, statreg, inst);
      sval = getword (sr, st, FALSE);
      val = sval - 2;
CHECK_STATUS:
      if ((sval & 0x8000) != (val & 0x8000))
      {
	 SET_CARRY;
	 SET_OVER;
      }
      else
      {
	 CLR_CARRY;
	 CLR_OVER;
      }
SAVE_CHECK_ZERO:
      putword (sr, st, val);
      cmpzero (val);
      break;

   case 10: /* BL */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X BL",
	       pcreg-2, wpreg, statreg, inst);
      if (st == 0)
      {
	 if (traceflag)
	    fprintf (tracefd, " R%d", sr);
	 val = GETREG (sr);
      }
      else val = getaddr (sr, st);
      PUTREG (11, pcreg);
      pcreg = val & 0xFFFE;
      break;

   case 11: /* SWPB */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X SWPB",
	       pcreg-2, wpreg, statreg, inst);
      val = getword (sr, st, FALSE);
      sval = val >> 8;
      sval |= (val & 0xFF) << 8;
      putword (sr, st, sval);
      break;

   case 12: /* SETO */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X SETO",
	       pcreg-2, wpreg, statreg, inst);
      putword (sr, st, 0xFFFF);
      break;

   case 13: /* ABS */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X ABS",
	       pcreg-2, wpreg, statreg, inst);
      val = getword (sr, st, FALSE);
      cmpzero (val);
      if (val == 0x8000) SET_OVER;
      else CLR_OVER;
      val = abs((int16)val);
      putword (sr, st, val);
      break;

   default:
      return (-1);
   }
   return (0);
}

/***********************************************************************
* proc7op - Process type 7 operations.
***********************************************************************/

static int
proc7op (uint16 inst)
{
   uint16 val;

   switch ((inst & 0x00F0) >> 4)
   {
   case 4: /* IDLE */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X IDLE",
	       pcreg-2, wpreg, statreg, inst);
      if (model >= 10 && IS_NONPRV)
      {
      }
      idle = TRUE;
      if ((statreg & 0x000F) == 0) run = FALSE;
      instcount--;
      pcreg -= 2;
      break;

   case 6: /* RSET */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X RSET",
	       pcreg-2, wpreg, statreg, inst);
      if (model >= 10 && IS_NONPRV)
      {
      }
      SET_MASK(0);
      stopclk();
      break;

   case 8: /* RTWP */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X RTWP",
	       pcreg-2, wpreg, statreg, inst);
      if (model >= 10 && IS_NONPRV)
      {
      }
      pcreg = R14 & 0xFFFE;
      statreg = R15;
      wpreg = R13 & 0xFFFE;
      deferint = 1;
      break;

   case 10: /* CKON */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X CKON",
	       pcreg-2, wpreg, statreg, inst);
      if (model >= 10 && IS_NONPRV)
      {
      }
      startclk();
      break;

   case 12: /* CKOF */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X CKOF",
	       pcreg-2, wpreg, statreg, inst);
      if (model >= 10 && IS_NONPRV)
      {
      }
      stopclk();
      break;

   case 14: /* LREX */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X LREX",
	       pcreg-2, wpreg, statreg, inst);
      if (model >= 10 && IS_NONPRV)
      {
      }
      val = wpreg;
      wpreg = GETMEM (0xFFFC) & 0xFFFE;
      PUTREG (13, val);
      PUTREG (14, pcreg);
      PUTREG (15, statreg);
      pcreg = GETMEM (0xFFFE) & 0xFFFE;
      SET_MASK(0);
      break;

   default:
      return (-1);
   }
   return (0);
}

/***********************************************************************
* proc8op - Process type 8 operations.
***********************************************************************/

static int
proc8op (uint16 inst)
{
   int32 acc;
   uint16 r;
   uint16 val;
   uint16 cval;
   int32 sv1, sv2;

   r = inst & 0x000F;

   if (inst == 0x0300) /* LIMI */
   {
      if (model >= 10 && IS_NONPRV)
      {
      }
      SET_MASK (GETINST);
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X LIMI %d",
	       pcreg-2, wpreg, statreg, inst, statreg & 0xF);
      pcreg += 2;
   }
   else switch ((inst & 0x00F0) >> 4)
   {
   case 0: /* LI */
      val = GETINST;
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X LI R%d %4X",
	       pcreg-2, wpreg, statreg, inst, r, val);
      pcreg += 2;
      PUTREG (r, val);
      cmpzero (val);
      break;

   case 2: /* AI */
      sv1 = GETINST;
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X AI R%d %4X",
	       pcreg-2, wpreg, statreg, inst, r, sv1);
      pcreg += 2;
      acc = sv1;
      sv2 = GETREG (r);
      acc = acc + sv2;
      if (acc & 0xFFFF0000) SET_CARRY;
      else CLR_CARRY;
      if (((sv1 & 0x8000) == (sv2 & 0x8000)) && (acc & 0x8000 != sv1 &0x8000))
         SET_OVER;
      else CLR_OVER;
      val = acc & 0xFFFF;
      PUTREG (r, val);
      cmpzero (val);
      break;

   case 4: /* ANDI */
      val = GETINST;
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X ANDI R%d %4X",
	       pcreg-2, wpreg, statreg, inst, r, val);
      val = GETREG(r) & val;
      pcreg += 2;
      PUTREG (r, val);
      cmpzero (val);
      break;

   case 6: /* ORI */
      val = GETINST;
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X ORI R%d %4X",
	       pcreg-2, wpreg, statreg, inst, r, val);
      val = GETREG(r) | val;
      pcreg += 2;
      PUTREG (r, val);
      cmpzero (val);
      break;

   case 8: /* CI */
      val = GETREG (r);
      cval = GETINST; pcreg += 2;
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X CI R%d %4X",
	       pcreg-2, wpreg, statreg, inst, r, cval);
      if (val > cval) SET_LGT;
      else CLR_LGT;
      if ((int16)val > (int16)cval) SET_AGT;
      else CLR_AGT;
      if (val == cval) SET_EQ;
      else CLR_EQ;
      break;

   case 10: /* STWP */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X STWP R%d",
	       pcreg-2, wpreg, statreg, inst, r);
      PUTREG (r, wpreg);
      break;

   case 12: /* STST */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X STST R%d",
	       pcreg-2, wpreg, statreg, inst, r);
      PUTREG (r, statreg);
      break;

   case 14: /* LWPI */
      wpreg = GETINST & 0xFFFE;
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X LWPI %4X",
	       pcreg-2, wpreg, statreg, inst, wpreg);
      pcreg += 2;
      break;

   default:
      return (-1);
   }
   return (0);
}

/***********************************************************************
* proc9op - Process type 9 operations.
***********************************************************************/

static int
proc9op (uint16 inst)
{
   int32 acc;
   int16 val;
   uint16 st;
   uint16 sr;
   uint16 r;
   uint16 uval;
   uint16 op;

   sr =  inst & 0x000F;
   st = (inst & 0x0030) >> 4;
   r = (inst & 0x03C0) >> 6;
   op = (inst & 0x0400) >> 10;

   switch ((inst & 0x0400) >> 10)
   {
   case 0: /* MPY */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X MPY ",
	       pcreg-2, wpreg, statreg, inst);
      val = getword (sr, st, TRUE);
      if (traceflag)
	 fprintf (tracefd, " R%d", sr);
      acc = GETREG (r);
      acc = acc * val;
      uval = (acc >> 16) & 0xFFFF;
      PUTREG (r, uval);
      uval = acc & 0xFFFF;
      PUTREG (r+1, uval);
      break;

   case 1: /* DIV */
      if (traceflag)
	 fprintf (tracefd, "PC %04X WP %04X ST %04X INST %04X DIV",
	       pcreg-2, wpreg, statreg, inst);
      val = getword (sr, st, TRUE);
      if (traceflag)
	 fprintf (tracefd, " R%d", sr);
      if ((int16)val <= (int16)GETREG(r))
      {
         SET_OVER;
      }
      else
      {
	 CLR_OVER;
	 acc = GETREG (r);
	 acc = (acc << 16 | GETREG (r+1));
	 PUTREG (r, acc / val);
	 PUTREG (r+1, acc % val);
      }
      break;

   default:
      return (-1);
   }
   return (0);
}

/***********************************************************************
* runinst - Run an instruction.
***********************************************************************/

static void 
runinst (uint16 inst)
{
   int type;

   type = decodeinst (inst);

#ifdef DEBUGINSTR
   screenposition (16,1);
   clearline();
   printf ("INST = %04X TYPE = %d", inst, type);
#endif

   switch (type)
   {
   case TYPE_1:
      if (proc1op (inst) < 0) goto ILL_INST;
      break;

   case TYPE_2:
      if (proc2op (inst) < 0) goto ILL_INST;
      break;

   case TYPE_3:
      if (proc3op (inst) < 0) goto ILL_INST;
      break;

   case TYPE_4:
      if (proc4op (inst) < 0) goto ILL_INST;
      break;

   case TYPE_5:
      if (proc5op (inst) < 0) goto ILL_INST;
      break;

   case TYPE_6:
      if (proc6op (inst) < 0) goto ILL_INST;
      break;

   case TYPE_7:
      if (proc7op (inst) < 0) goto ILL_INST;
      break;

   case TYPE_8:
      if (proc8op (inst) < 0) goto ILL_INST;
      break;

   case TYPE_9:
      if (proc9op (inst) < 0) goto ILL_INST;
      break;

   default:
ILL_INST:
      sprintf (view[0], "Illegal instruction: inst = %04X, pc = %04X",
	       inst, pcreg-2);
      idle = FALSE;
      run = FALSE;
      return;
   }
   instcount++;
   if (traceflag)
      fprintf (tracefd, "\n");
}

/***********************************************************************
* stepprog - Step through a program.
***********************************************************************/

void 
stepprog (void)
{
   int type;
   uint16 inst;

   inst = GETINST;
   pcreg += 2;
   pcreg &= 0xFFFE;
   runinst (inst);
}

/***********************************************************************
* runprog - Run a program.
***********************************************************************/

void
runprog (void)
{
   ttrunstate();

   while (run)
   {
      idle = FALSE;
      stepprog();
      if (breakflag && pcreg == breakaddr)
      {
         run = FALSE;
      }
      else
      {
	 if (enablepanel && pancount) panel();
	 if (statreg & 0x000F)
	 {
	    if (deferint == 0)
	       checkinterrupts();
	    else deferint--;
	 }
      }
   }

   ttcmdstate();
}
